/*
	File:		ListEx1PartUtils.cpp

	Contains:	ListEx1Part utility functions & classes

	Written by:	Steve Smith

	Copyright:	 1994,95 by Apple Computer, Inc., all rights reserved.
*/

// -- Compiler/Preprocessor Switches --

#ifndef _COMPILERDEFS_
#include "CompDefs.h"
#endif

// -- OpenDoc Utilities --

#ifndef _EXCEPT_
// Exceptions define several important macros (ie. CHECKENV)
// which are used in the SOM method dispatch glue. If Except.h
// is not included early enough, exceptions may not be thrown
// correctly when returning from a SOM method with the "ev" parameter set.
#include <Except.h>
#endif

// --- ListEx1Part Includes ---

#ifndef _LISTEX1PARTUTILS_
#include "ListEx1PartUtils.h"
#endif

#ifndef _LISTEX1PARTDEF_
#include "ListEx1PartDef.h"
#endif

// --- OpenDoc Includes ---

#ifndef _ODTYPES_
#include <ODTypes.h>
#endif

#ifndef SOM_Module_OpenDoc_StdProps_defined
#include <StdProps.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdTypes_defined
#include <StdTypes.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdDefs_defined
#include <StdDefs.xh>
#endif

#ifndef SOM_ODFrame_xh
#include <Frame.xh>
#endif

#ifndef SOM_ODSession_xh
#include <ODSessn.xh>
#endif

#ifndef SOM_ODDraft_xh
#include <Draft.xh>
#endif

#ifndef SOM_ODStorageUnit_xh
#include <StorageU.xh>
#endif

#ifndef SOM_ODStorageUnitView_xh
#include <SUView.xh>
#endif

#ifndef SOM_ODDocument_xh
#include <Document.xh>
#endif

#ifndef SOM_ODContainer_xh
#include <ODCtr.xh>
#endif

#ifndef SOM_ODNameSpaceManager_xh
#include <NmSpcMg.xh>
#endif

#ifndef SOM_ODValueNameSpace_xh
#include <ValueNS.xh>
#endif

// -- OpenDoc Utilities --

#ifndef _BNDNSUTL_
#include <BndNSUtl.h>
#endif

#ifndef _DOCUTILS_
#include <DocUtils.h>
#endif

#ifndef _ISOSTR_
#include <ISOStr.h>
#endif

#ifndef _ODDEBUG_
#include <ODDebug.h>
#endif

#ifndef _STDTYPIO_
#include <StdTypIO.h>
#endif

#ifndef _STORUTIL_	
#include <StorUtil.h>
#endif

#ifndef _TEMPOBJ_
#include <TempObj.h>
#endif

#ifndef _TEMPITER_
#include <TempIter.h>
#endif

#ifndef _USERSRCM_
#include <UseRsrcM.h>
#endif

// --- Macintosh Includes ---

#ifndef __RESOURCES__
#include <Resources.h>
#endif

#ifndef __GXMATH__
#include <GXMath.h>
#endif

#ifndef __SCRIPT__
#include <Script.h>
#endif

#ifndef __TEXTUTILS__
#include <TextUtils.h>
#endif


#pragma segment ListEx1PartUtilities


//====================================================================
// Utility Functions
//====================================================================

//--------------------------------------------------------------------
// GetPartName
//--------------------------------------------------------------------

ODIText* GetPartName(Environment* ev, ODPart* part, ODType category)
{
	ASSERT(part != kODNULL, kODErrIllegalNullPartInput);

	// Get the root part of the document.
	ODDraft* draft = ODGetDraft(ev,part);
	TempODPart rootPart = ODAcquireRootPartOfDraft(ev, draft);
	
	// Get the part name.
	// If we can't get a valid name for the part, we generate one using
	// the user visible category name of the part's category.
	TempODIText partName = ODGetITextProp(ev, part->GetStorageUnit(ev), 
										 kODPropName, kODMacIText, kODNULL);
	
	// If the part is the root of the document, return the name of the file.
	if ( ODObjectsAreEqual(ev, part, rootPart) )
	{
		ODContainer* container = draft->GetDocument(ev)->GetContainer(ev);
		TempPlatformFile file = GetPlatformFileFromContainer(ev, container);

		TempODIText fileName = file->GetName();
		
		// Test the file name against the part name. If the two are equivalent
		// (not equal), then use the part name.
		if ( !NamesAreEquivalent(ev, fileName, partName) )
		{
			// If the names are different, return the file name.
			// (This code transfers the ODIText object to the partName tempobj)
			DisposeIText(partName.DontDelete());
			partName = fileName.DontDelete();
		}
	}
	else
	{
		if ( (partName == kODNULL) || (GetITextStringLength(partName) == 0) )
		{
			ODIText* categoryName;
			ODNameSpaceManager* nsMgr = ODGetSession(ev,part)->GetNameSpaceManager(ev);
			
			// Get the category string from the category name space.
			if ( GetUserCatFromCat(nsMgr, category, &categoryName) )
			{
				// If we successfully retrieved the category user string, return it.
				// (This code transfers the ODIText object to the partName tempobj)
				DisposeIText(partName.DontDelete());
				partName = categoryName;
			}
			else
			{
	#if ODDebug
				// This should never happen. Check NMAP for errors.
				DebugStr("\pCategory NMAP bad, or Preferences corrupted.");
	#else
				THROW(kODErrInvalidNSType);
	#endif
			}
		}
	}
	
	return partName.DontDelete();
}

//--------------------------------------------------------------------
// NamesAreEquivalent
//--------------------------------------------------------------------

ODBoolean NamesAreEquivalent(Environment* ev, ODIText* fileName,
										ODIText* partName)
{
	Str255 fileStr;
	Str255 partStr;
	
	// If the strings are in different languages, we're done.
	if ( (GetITextScriptCode(fileName) != GetITextScriptCode(partName)) ||
			GetITextLangCode(fileName) != GetITextLangCode(partName) )
		return kODFalse;
	
	GetITextPString(fileName, fileStr);
	GetITextPString(partName, partStr);

	// If the string lengths are different, we're done.
	if ( fileStr[0] != partStr[0] )
		return kODFalse;

	// Return the Toolbox string equivalence test.
	return EqualString(fileStr,partStr,kODTrue,kODTrue);
}

//--------------------------------------------------------------------
// GetEditorScriptLanguage
//--------------------------------------------------------------------

void GetEditorScriptLanguage(Environment* ev, ODScriptCode* script,
										ODLangCode* language)
{
	long region;
	
	ODSLong rfRef;
	rfRef = BeginUsingLibraryResources();
	{
		Handle versHdl = Get1Resource('vers', 1);
		
		// Get the region code of the editor, otherwise use the
		// region code the of the primary system script.
		
		if ( versHdl )
		{
			region = (long)(*(VersRecHndl)versHdl)->countryCode;
			ReleaseResource(versHdl);
		}
		else
		{
			region = GetScriptManagerVariable(smRegionCode);
		}
		
		// Spanish & Japanese are not actually supported by the editor.
		// They are provided as examples of how to add recognition of
		// additional regions (see Script.h for region codes).
		
		switch ((short)region) {
			case verUS:
				*script = smRoman;
				*language = langEnglish;
				break;
			case verSpain:
				*script = smRoman;
				*language = langSpanish;
				break;
			case verJapan:
				*script = smJapanese;
				*language = langJapanese;
				break;
			default:
				*script = smRoman;
				*language = langEnglish;
		}				
	}
	EndUsingLibraryResources(rfRef);
}

//--------------------------------------------------------------------
// FixedToIntRect
//--------------------------------------------------------------------

void FixedToIntRect(ODRect& fixedRect, Rect& intRect)
{
	intRect.top		= FixedToInt(fixedRect.top);
	intRect.left	= FixedToInt(fixedRect.left);
	intRect.bottom	= FixedToInt(fixedRect.bottom);
	intRect.right	= FixedToInt(fixedRect.right);
}

//--------------------------------------------------------------------
// IntToFixedRect
//--------------------------------------------------------------------

void IntToFixedRect(Rect& intRect, ODRect& fixedRect)
{
	fixedRect.left		= ff(intRect.left);
	fixedRect.top		= ff(intRect.top);
	fixedRect.right		= ff(intRect.right);
	fixedRect.bottom	= ff(intRect.bottom);
}

//--------------------------------------------------------------------
// LoadThumbnail
//--------------------------------------------------------------------

void LoadThumbnail(Environment* ev, Handle* thumbnail)
{
	if ( *thumbnail ) return;
	
	ODSLong rfRef;
	rfRef = BeginUsingLibraryResources();
	{
		*thumbnail = (Handle) GetPicture(kThumbnailPicture);
	}
	EndUsingLibraryResources(rfRef);
}

//--------------------------------------------------------------------
// TilePartWindow
//--------------------------------------------------------------------

Rect TilePartWindow(Environment* ev, Rect* facetBounds, Rect* partWindowBounds)
{
	const short	kWindowTilingConst	= 20;
	const short kLeftToRight		= 0;
	const short kRightToLeft		= -1;
	
	short direction;
	
	// Get the direction for the primary script system running on this machine.
	// (Right-to-Left or Left-to-Right)
	direction = GetSysDirection();
	
	// The child window should be tiled from the topLeft corner of the 
	// active facet whose frame is being opened.
	if ( direction == kLeftToRight )
	{
		// Position the window rect at the top/left corner of the facet.
		OffsetRect(partWindowBounds, facetBounds->left, facetBounds->top);
		// Now tile the window rect down and to the right.
		OffsetRect(partWindowBounds, kWindowTilingConst, kWindowTilingConst);
	}
	// The child window should be tiled from the topRight corner of the 
	// active facet whose frame is being opened.
	else if ( direction == kRightToLeft )
	{
		// Position the window rect at the top/right corner of the facet.
		OffsetRect(partWindowBounds, (partWindowBounds->right - facetBounds->right),
					facetBounds->top);
		// Now tile the window rect down and to the left.
		OffsetRect(partWindowBounds, -kWindowTilingConst, kWindowTilingConst);
	}
	
	return *partWindowBounds;
}

//--------------------------------------------------------------------
// CountFacets
//--------------------------------------------------------------------

ODUShort CountFramesFacets(Environment* ev, ODFrame* frame)
{
	ODUShort facetCount = 0;
	ODFacet* facet;
	
	TempODFrameFacetIterator ffiter(ev, frame);
	facet = ffiter.First();
	while ( ffiter.IsNotComplete() )
	{
		facetCount++;
		facet = ffiter.Next();
	}
	
	return facetCount;
}

//=========================================================================
// CFrameProxy
//=========================================================================

//-------------------------------------------------------------------------
// CFrameProxy::SetFrame
//
// The code will not affect the proxy fields unless it can
// successfully acquire the incoming frame, its id, and
// the draft the frame lives in. If something goes wrong,
// the proxy remains unchanged.
//-------------------------------------------------------------------------

void CFrameProxy::SetFrame(Environment* ev, ODFrame* frame)
{
	ASSERT(frame!=kODNULL, kODErrIllegalNullFrameInput);
	
	ODID id;
	ODDraft* draft;
	
	// Using the temp object will cleanup the incoming
	// frame's refcount if something goes wrong.
	frame->Acquire(ev);
	TempODFrame tFrame = frame;
	
	// Perform all the operations that can fail, first.
	TempODPart tPart = frame->AcquirePart(ev);
	id = frame->GetID(ev);
	draft = ODGetDraft(ev, tPart);
	ODReleaseObject(ev, fFrame);

	// If all went well, set the proxy fields.
	fFrame = tFrame.DontRelease();
	fID = id;
	fDraft = draft;
}

//-------------------------------------------------------------------------
// CFrameProxy::GetFrame
//-------------------------------------------------------------------------

ODFrame* CFrameProxy::GetFrame(Environment* ev)
{
	if ( fFrame == kODNULL )
	{
		TRY
			TempODFrame frame = fDraft->AcquireFrame(ev, fID);
			this->SetFrame(ev,frame);
		CATCH_ALL
			fFrame = kODNULL;
			fID = kODNULLID;
		ENDTRY
	}
	return fFrame;
}

//-------------------------------------------------------------------------
// CFrameProxy::Purge
//-------------------------------------------------------------------------

void CFrameProxy::Purge(Environment* ev)
{
	if ( this->FrameIsLoaded() )
	{
		fID = fFrame->GetID(ev);
		ODReleaseObject(ev, fFrame);
	}
}

//=========================================================================
// CFrameInfo
//=========================================================================

//-------------------------------------------------------------------------
// CFrameInfo::CFrameInfo
//-------------------------------------------------------------------------

CFrameInfo::CFrameInfo()
{
	fFrameActive = kODFalse;
	fFrameReactivate  = kODFalse;
	fShouldDisposeWindow = kODFalse;
	fActiveFacet = kODNULL;
	fSourceFrame = kODNULL;
	fDependentFrame = kODNULL;
	fPartWindow = kODNULL; 
}

//-------------------------------------------------------------------------
// CFrameInfo::~CFrameInfo
//-------------------------------------------------------------------------

CFrameInfo::~CFrameInfo()
{
	// Release the part window, if we still have a reference to it.
	ODSafeReleaseObject(fPartWindow); 
	// Deleting the proxies will release the encapsulated frames.
	ODDeleteObject(fDependentFrame);
	ODDeleteObject(fSourceFrame);
}

//-------------------------------------------------------------------------
// CFrameInfo::Externalize
//-------------------------------------------------------------------------

void CFrameInfo::Externalize(Environment* ev, ODStorageUnitView* storageUnitView)
{
	// This method assumes that OpenDoc has passed us a storageUnitView
	// that is focused to a property, but no particular value.
	
	ODStorageUnit* storageUnit = storageUnitView->GetStorageUnit(ev);

	this->CleanseFrameInfoProperty(ev, storageUnit);
	this->ExternalizeFrameInfo(ev, storageUnit, kODNULLKey, kODNULL);
}

//-------------------------------------------------------------------------
// CFrameInfo::CleanseFrameInfoProperty
//-------------------------------------------------------------------------

void CFrameInfo::CleanseFrameInfoProperty(Environment* ev, ODStorageUnit* storageUnit)
{
	ODULong numValues;
	ODULong index;
		
	numValues = storageUnit->CountValues(ev);
	
	for (index = numValues; index >= 1; index--)
	{
		// Index, from 1 to n, through the values.
		storageUnit->Focus(ev, kODNULL, kODPosSame, 
								kODNULL, index, kODPosUndefined);
								
		// Get the ISO type name for the value. The temp object
		// will automatically delete the returned value when this
		// scope is exited.
		TempODValueType value = storageUnit->GetType(ev);
		
		// If the value type is not one we support, remove it.
		if ( ODISOStrCompare(value, kListEx1PartInfo) != 0 )
			storageUnit->Remove(ev);
	}
}

//-------------------------------------------------------------------------
// CFrameInfo::ExternalizeFrameInfo
//-------------------------------------------------------------------------

void CFrameInfo::ExternalizeFrameInfo(Environment* ev, ODStorageUnit* storageUnit,
										ODDraftKey key, ODFrame* scopeFrame)
{
	// This method behaves much the same way as the ListEx1Part::ExternalizeStateInfo
	// method.
	
	if ( storageUnit->Exists(ev, kODNULL, kListEx1PartInfo, 0) )
	{
		// Persistent object references are stored in a side table, rather than
		// in the property/value stream. Thus, deleting the contents of a value
		// will not "delete" the references previously written to that value. To
		// completely "delete" all references written to the value, we must
		// remove the value and add it back.

		storageUnit->Focus(ev, kODNULL, kODPosSame, kListEx1PartInfo, 0, kODPosUndefined);
		storageUnit->Remove(ev);
	}

	// Add a value to write the data into.
	storageUnit->AddValue(ev, kListEx1PartInfo);
	
	// Write a weak reference to our source frame.
	{
		ODStorageUnitRef weakRef = {0,0,0,0};
		
		if ( fSourceFrame )
		{
			ODID		frameID = fSourceFrame->GetID();
			ODID		scopeFrameID = ( scopeFrame ? scopeFrame->GetID(ev) : kODNULLID );
			ODDraft*	fromDraft = fSourceFrame->GetDraft();
	
			// If a draft key exists, then we are being cloned to another draft.
			// We must "weak" clone our display frame and reference the cloned
			// frame. The part re-uses the frameID variable so there aren't two
			// different GetWeakStorageUnitRef calls.
			if ( key )
				frameID = fromDraft->WeakClone(ev, key, frameID, kODNULLID, scopeFrameID);
			
			// Write out weak references to each of the part's display frames.
			storageUnit->GetWeakStorageUnitRef(ev, frameID, weakRef);
		}
		StorageUnitSetValue(storageUnit, ev, sizeof(ODStorageUnitRef), (ODPtr)&weakRef);
	}
	
	// Write a weak reference to our dependent frame.
	{
		ODStorageUnitRef weakRef = {0,0,0,0};
		
		if ( fDependentFrame )
		{
			ODID		frameID = fDependentFrame->GetID();
			ODID		scopeFrameID = ( scopeFrame ? scopeFrame->GetID(ev) : kODNULLID );
			ODDraft*	fromDraft = fDependentFrame->GetDraft();
	
			// If a draft key exists, then we are being cloned to another draft.
			// We must "weak" clone our display frame and reference the cloned
			// frame. The part re-uses the frameID variable so there aren't two
			// different GetWeakStorageUnitRef calls.
			if ( key )
				frameID = fromDraft->WeakClone(ev, key, frameID, kODNULLID, scopeFrameID);
			
			// Write out weak references to each of the part's display frames.
			storageUnit->GetWeakStorageUnitRef(ev, frameID, weakRef);
		}
		StorageUnitSetValue(storageUnit, ev, sizeof(ODStorageUnitRef), (ODPtr)&weakRef);
	}
}

//-------------------------------------------------------------------------
// CFrameInfo::CloneInto
//-------------------------------------------------------------------------

void CFrameInfo::CloneInto(Environment *ev, ODDraftKey key,
							ODStorageUnitView* storageUnitView, ODFrame* scopeFrame)
{
	// This method assumes that OpenDoc has passed us a storageUnitView
	// that is focused to a property, but no particular value.
	
	ODStorageUnit* storageUnit = storageUnitView->GetStorageUnit(ev);

	if ( storageUnit->Exists(ev, kODNULL, kListEx1PartInfo, 0) == kODFalse )
	{
		this->ExternalizeFrameInfo(ev, storageUnit, key, scopeFrame);
	}
}

//-------------------------------------------------------------------------
// CFrameInfo::InitFromStorage
//------------------------------------------------------------------------------

void CFrameInfo::InitFromStorage(Environment* ev, ODStorageUnitView* storageUnitView)
{
	// This method assumes that OpenDoc has passed us a storageUnitView
	// that is focused to a property, but no particular value.
	
	ODStorageUnit* storageUnit = storageUnitView->GetStorageUnit(ev);

	if ( storageUnit->Exists(ev, kODNULL, kListEx1PartInfo, 0) )
	{
		TRY
			storageUnit->Focus(ev, kODNULL, kODPosSame,
										kListEx1PartInfo, 0 , kODPosUndefined);
	
			ODStorageUnitRef weakRef = {0,0,0,0};
			StorageUnitGetValue(storageUnit, ev, sizeof(ODStorageUnitRef),
								(ODPtr)&weakRef);
			
			if ( storageUnit->IsValidStorageUnitRef(ev, weakRef) )
			{
				// Convert the reference into a runtime id.
				ODID frameID = storageUnit->GetIDFromStorageUnitRef(ev, weakRef);

				// Create a proxy class to support the lazy internalization.
				CFrameProxy* proxy = new CFrameProxy;
				proxy->InitFrameProxy(frameID, ODGetDraft(ev,storageUnit));

				// Store the proxy source frame.
				fSourceFrame = proxy;
			}
			
		CATCH_ALL
			ODDeleteObject(fSourceFrame);
			fSourceFrame = kODNULL;
		ENDTRY

		TRY
			ODStorageUnitRef weakRef = {0,0,0,0};
			StorageUnitGetValue(storageUnit, ev, sizeof(ODStorageUnitRef),
								(ODPtr)&weakRef);
			
			if ( storageUnit->IsValidStorageUnitRef(ev, weakRef) )
			{
				// Convert the reference into a runtime id.
				ODID frameID = storageUnit->GetIDFromStorageUnitRef(ev, weakRef);

				// Create a proxy class to support the lazy internalization.
				CFrameProxy* proxy = new CFrameProxy;
				proxy->InitFrameProxy(frameID, ODGetDraft(ev,storageUnit));

				// Store the proxy dependent frame.
				fDependentFrame = proxy;
			}
			
		CATCH_ALL
			ODDeleteObject(fDependentFrame);
			fDependentFrame = kODNULL;
		ENDTRY
	}
}

//-------------------------------------------------------------------------
// CFrameInfo::SetSourceFrame
//-------------------------------------------------------------------------

void CFrameInfo::SetSourceFrame(Environment* ev, ODFrame* frame)
{
	if ( frame != kODNULL )
	{
		// Create a proxy class to support the lazy internalization.
		CFrameProxy* proxy = new CFrameProxy;
		proxy->InitFrameProxy(ev,frame);
		
		// Store the proxy source frame after clearing the old one.
		ODDeleteObject(fSourceFrame);
		fSourceFrame = proxy;
	}
}

//-------------------------------------------------------------------------
// CFrameInfo::ReleaseSourceFrame
//-------------------------------------------------------------------------

void CFrameInfo::ReleaseSourceFrame(Environment* ev)
{
	ODDeleteObject(fSourceFrame);
}

//-------------------------------------------------------------------------
// CFrameInfo::SetDependentFrame
//-------------------------------------------------------------------------

void CFrameInfo::SetDependentFrame(Environment* ev, ODFrame* frame)
{
	if ( frame != kODNULL )
	{
		// Create a proxy class to support the lazy internalization.
		CFrameProxy* proxy = new CFrameProxy;
		proxy->InitFrameProxy(ev,frame);
		
		// Store the proxy dependent frame after clearing the old one.
		ODDeleteObject(fDependentFrame);
		fDependentFrame = proxy;
	}
}

//-------------------------------------------------------------------------
// CFrameInfo::ReleaseDependentFrame
//-------------------------------------------------------------------------

void CFrameInfo::ReleaseDependentFrame(Environment* ev)
{
	ODDeleteObject(fDependentFrame);
}





